/**
 * shelldom.js - A JavaScript Linux-like console by Stefano Adriani.
 *
 * The ShellDOM project is a set of JavaScript libraries, structured using
 * an Object-Oriented approach and the MVC pattern. Aims of the projects are:
 *
 * 1. To offer an easy, quick way to learn (online) Linux basic commands.
 * Target of this usage are web developers, HTML experts, and web designers which
 * are familiar with the DOM structure. The ShellDOM console should allow them 
 * to practice the basic GNU commands, while "moving" in a known enviroment (the DOM tree).
 *
 * 2. To offer an easy, quick way to learn on line the DOM structure.
 * Target of this usage are Linux administration, GNU experts, and linux developers which
 * are familiar with the GNU terminal. The ShellDOM console should allow them 
 * to learn the DOM structure, using a known set of instructions (the GNU console commands).
 * 
 * Basically the ShellDOM project could be considered a try to "mix together" and "move closer"
 * two different world and set of skills: the "DOM-WEB" world and the "LINUX-GNU" universe.
 *
 * No other library is required: this is just old plain JavaScript.
 *
 * Permission is hereby granted to use this version of the library under the
 * same terms as shelldom.js, which has the following license:
 *
 * --
 * Copyright (c) 2012 Stefano Adriani  (adriani.altervista.com)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * The Software shall be used for Good, not Evil.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * --
 *
 * @package ShellDOM
 * @author Stefano Adriani <adriani@altervista.com>
 * @copyright 2012 Stefano Adriani <adriani@altervista.com> (shelldom.js)
 * @copyright 2012 Stefano Adriani <adriani@altervista.com> (shelldomImpl.js)
 * @license http://opensource.org/licenses/mit-license.php MIT License
 * @version 1.0 (20-09-2012)
 * @link http://adriani.altervista.org/work/js/shelldom.php
 */

// The class for a single line (in the shell's view)
function LogLine(msg, index) {
	this.max		= 90 ;
	this.line 		= msg ;
	this.index		= index ;
	this.error		= false ;
	this.prompt		= "$&nbsp;" ;
	// this.prompt		= "C:\\>" ;
	// this.prompt		= this.index + ": " ;
	this.open_tag	= "<p class='log-line'>" ;
	this.line_start = this.open_tag + this.prompt ;
	this.line_error = "<p class='log-error'>" + this.prompt ;
	this.line_cmd	= "<p class='log-line' id='last_cmd'>" + this.prompt ;
	this.line_end 	= "</p>" ;
	this.reset		= function() {	
		this.line	= "" ;
		this.error	= false ;
	}
	this.getLenth	= function (s) {
		var plain = s.replace(/&nbsp;/g," ") ;
		return plain.length ;
	}
	// Return number of occurences of character "c" within string "s"
	this.count		= function(s, c) { 
		return (s.length - s.replace(new RegExp(c,"g"),'').length)/c.length ;
	}
	this.getLines	= function(value) {
		if (value == null)					return 0 ;
		else if (value.length == 0)			return 1 ;
		else if (value.length == undefined)	return 1 ;
		else {
			//console.log(value + " = " + Math.ceil(100*this.getLenth(value)/this.max)/100) ;
			return Math.ceil(this.getLenth(value)/this.max) ;
		}
	}
	this.getSize	= function() 		{ return this.getLines(this.line) ; }
	this.format		= function(value) 	{
		var result 	= "" ;
		var start 	= this.line_cmd ;
		var	command = (value != null ) ;
		if (!command) {
			value = this.line ;
			start = this.error ? this.line_error : this.line_start;
		}
		var	actual	= 0 ;
		var output	= new String(value) ;
		for (k = 0 ; k < this.getLines(value) ; k++) {
			actual = this.max + 4*this.count(value, "&nbsp;") ;
			output = value.substring(actual*k, actual*(k+1)) ;
			if (k == 0)	result = start + output + this.line_end ; 
			else		result += this.open_tag + output + this.line_end ; 
			//else		result += this.open_tag + "&nbsp;&nbsp;" + output + " #" + this.getLines(value) + "#" + this.line_end ; 
		}
		return result ;
	}
}
// The class for the SHELL MODEL
function ShellModel(view) {
	this.handler ;
	this.view		= view ;
	this.setHandler	= function(h) 	{ this.handler = h ; 		}
	this.getHandler	= function() 	{ return this.handler ; 	}
	this.runCommand	= function(cmd) {
		try 		{ eval(cmd) ; 								} 
		catch(err) 	{ this.view.show(err.message, 'green') ;	}	
	}
	this.echo		= function(arg) { this.view.scroll(arg) ; 	}
	this.clear		= function() {
		for (i = 0; i < this.view.max_lines ; i++) this.view.getLog()[i].reset() ;
		this.view.scroll("") ;
	}
	this.ls			= function(arg) { this.handler.ls(arg) ;	}	
	this.cd			= function(arg)	{ this.handler.cd(arg) ;	}	
	this.pwd		= function(cmd) { this.handler.pwd() ;		}	
	this.history	= function() {
		for (n = this.view.buffer.length-1 ; n >= 0 ; n--) this.view.scroll(this.view.buffer[n]) ;
	}
	this.help		= function() {
		this.view.scroll("shell DOM ver. 1.0") ;
		this.view.scroll("&copy; Stefano Adriani 2012") ;
		this.view.scroll("Click <a href='http://adriani.altervista.org/work/js/shelldom.php#help'>here</a> for help") ;
	}
	this.adams		= function() { this.view.scroll("Life, the Universe, and Everything") ;	}	
}
// The class for the SHELL VIEW
function ShellView(url, lines) {
	this.url 		= url;
	this.last_key	= 0 ;
	this.max_lines	= lines ;
	this.buffer 	= new Array(0) ;
	this.log 		= new Array(this.max_lines) ;
	this.model		= new ShellModel(this) ;
	this.getLog 	= function() { return this.log ;									}
	this.getModel 	= function() { return this.model ; 									}
	this.getPrompt 	= function() { return this.log[0].prompt ;					 		}
	this.getMonitor = function() { return document.getElementById("shell-monitor") ; 	}	
	this.getInput 	= function() { return document.getElementById("shell-input"); 		}
	this.init	 	= function() { for (i = 0; i < this.max_lines ; i++) this.log[i] = new LogLine("", i) ; }
	this.memoryUp	= function() {
		var cmd = this.getInput().value = this.buffer[this.last_key] ;
		if (this.last_key < (this.buffer.length-1)) this.last_key++ ;
	}
	this.memoryDown	= function() {
		if (this.last_key > 0) this.last_key-- ;
		var cmd = this.getInput().value = this.buffer[this.last_key] ;
	}
	this.scroll	= function(cmd) {
		var lines 	= 0 ;
		var delta	= 0 ;
		var i 		= this.max_lines ;
		monitor 	= this.getMonitor() ;
		monitor.innerHTML = "" ;
		// 1. Acknowledging how many log lines we can print, ensuring to print always the last line
		while (lines < this.max_lines) {
			i-- ;
			lines += this.log[i].getSize() ;
		}
		if (this.log[0].getLines(cmd) > 1) 	i += (this.log[0].getLines(cmd) - 1) ;
		if (this.log[i].getSize() <= 1) 	delta = 1 ;
		else {	
			delta = this.log[i].getSize() ;
			i -= (delta - 1) ; 
		}
		// 2. Printing all log lines 
		while (i < this.max_lines) {
			if (i < ( this.max_lines - 1) ) {
				this.log[i].error 	= false ;
				var next = ((i + delta) < this.max_lines) ? delta : 1 ;
				this.log[i].line 	= this.log[i+next].line ;
				if (this.log[i+next].error) this.log[i].error 	= true ;
				monitor.innerHTML += this.log[i].format() ;
			} else {
				this.log[i].line = cmd ;
				monitor.innerHTML += this.log[i].format(cmd) ;
				this.log[i].error = false ;
			}
			i++ ;
		}
	}
	this.print	= function(cmd)			{ this.scroll(cmd.toLowerCase()) ; }	
	this.show	= function(cmd, color) 	{
		this.scroll(cmd) ;
		this.log[this.max_lines - 1].error = true ;
		document.getElementById("last_cmd").style.color = color ;
	}
}
// The class for the SHELL CONTROLLER
function ShellController(view) {
	this.view		= view ;
	this.model		= view.getModel() ;
	this.setHandler	= function(h) { this.model.setHandler(h) ; }
	this.execute	= function(e) {
		if (e.keyCode == 13) {
			var input 	= this.view.getInput() ;
			var cmd		= input.value.split(" ")[0] ;
			var arg		= input.value.split(" ")[1] ;
			// Echoes current command
			this.view.scroll(input.value) ;
			// Managing long commands
			if (cmd.length >= 5) {
				var key = cmd.substring(0,5) ;
				switch(key) {
					case	"clear"	: this.model.clear() ;				break ;
					case	"histo"	: this.model.history() ;			break ;
					case	"credi"	: this.model.help() ;				break ;
					default 		: this.model.runCommand(cmd) ;
				}
			// Managing 4 characters length
			} else if (cmd.length == 4) {
				switch(cmd) {
					case	"echo"	: this.model.echo(arg) ;			break ;
					case	"help"	: this.model.help() ;				break ;
					case	"info"	: this.model.help() ;				break ;
					default 		: this.model.runCommand(cmd.value) ;
				}
			// Managing 3 characters length
			} else if (cmd.length == 3) {
				switch(cmd) {
					case "pwd"		: this.model.pwd() ;			break ;
					case "cls"		: this.model.clear() ;			break ;
					case "man"		: this.model.help() ;			break ;
					case "ver"		: this.model.help() ;			break ;
					default 		: this.model.runCommand(cmd) ;
				}
			// Managing 2 characters length
			} else if (cmd.length == 2) {
				switch(cmd) {
					case "ll"		: this.model.ls("-l") ;			break ;
					case "ls"		: this.model.ls(arg) ;			break ;
					case "cd"		: this.model.cd(arg) ;			break ;
					case "42"		: this.model.adams() ;			break ;
					default 		: this.model.runCommand(cmd) ;
				}
			} else {
				if (cmd.length > 0) this.model.runCommand(cmd)	;
			}
			// Storing command in the buffer
			if (cmd.length > 0) this.view.buffer.unshift(input.value) ;
			input.value = "" ;
			return false ;
		} 
		else if (e.keyCode == 38) this.view.memoryUp() ;
		else if (e.keyCode == 40) this.view.memoryDown() ;
		//console.log("e.keyCode = " + e.keyCode + "-" + e.which + "-" + e.charCode) ;
	}
}