/**
 * shelldomImpl.js - Demo implementation of ShellDOM console by Stefano Adriani.
 *
 * The ShellDOM project is a set of JavaScript libraries, structured using
 * an Object-Oriented approach and the MVC pattern. Aims of the projects are:
 *
 * 1. To offer an easy, quick way to learn (online) the Linux basic commands.
 * Target of this usage are web developers, HTML experts, and web designers which
 * are familiar with the DOM structure. The ShellDOM console should allow them 
 * to practice the basic GNU commands, while "moving" in a known enviroment (the DOM tree).
 *
 * 2. To offer an easy, quick way to learn (online) the DOM structure.
 * Target of this usage are Linux administration, GNU experts, and linux developers which
 * are familiar with the GNU terminal. The ShellDOM console should allow them 
 * to learn the DOM structure, using a known set of instructions (the GNU console commands).
 * 
 * Basically the ShellDOM project could be considered a try to "mix together" and "move closer"
 * two different world and set of skills: the "DOM-WEB" world and the "LINUX-GNU" universe.
 *
 * No other library is required: this is just old plain JavaScript.
 *
 * Remark: due to "the Same origin policy for JavaScript", it's not possible to make cross-domain
 * Ajax calls. For details see:
 * https://developer.mozilla.org/en-US/docs/Same_origin_policy_for_JavaScript
 * http://fettig.net/weblog/2005/11/28/how-to-make-xmlhttprequest-connections-to-another-server-in-your-domain/
 *
 * Permission is hereby granted to use this version of the library under the
 * same terms as shelldom.js, which has the following license:
 *
 * ---
 * Copyright (c) 2012 Stefano Adriani  (adriani.altervista.com)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * The Software shall be used for Good, not Evil.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * ---
 *
 * @package ShellDOM
 * @author Stefano Adriani <adriani@altervista.com>
 * @copyright 2012 Stefano Adriani <adriani@altervista.com> (shelldomImpl.js)
 * @copyright 2012 Stefano Adriani <adriani@altervista.com> (shelldom.js)
 * @license http://opensource.org/licenses/mit-license.php MIT License
 * @version 1.0 (20-09-2012)
 * @link http://adriani.altervista.org/work/js/shelldom.php
 */

// The class for the TAG MANAGER
function TagManager() {
	this.tag 		= new Array(0) ;
	this.amount		= new Array(0) ;
	this.reset		= function() 	{ 
		this.tag 	= new Array(0) ; 
		this.amount = new Array(0) ; 
	}
	// Identify, count and assign number to siblings
	this.format		= function(e) 	{
		var found	= -1 ;
		for (i = 0 ; i < this.tag.length ; i++) if (this.tag[i] == e.tagName) found = i ;
		if (found < 0) {
			this.tag.push(e.tagName) ;
			this.amount.push(1) ;
			return e.tagName ;
		} else {
			this.amount[found]++ ;
			return e.tagName + this.amount[found] ;
		}
	}
	this.debug		= function() 	{ for (i = 0 ; i < this.tag.length ; i++) console.log(this.tag[i] + " - " + this.amount[i]) ; }
}
// The class for the DOM HANDLER
function HandlerDOM(view) {
	this.home ;
	this.current ;
	this.previous ;
	this.blank		= "&nbsp;" ;
	this.view		= view ;
	this.manager	= new TagManager() ;
	this.setHome	= function(node) { 
		this.home 		= node ;
		this.current	= this.home ;
		this.previous	= this.home ;
	}
	this.getHome	= function() 	 { return this.home ; }
	this.getPath	= function(node) {
		var path 	= [node.tagName] ;
		var parent 	= node.parentElement || node.parentNode ;
		while (parent) {
			if (parent.tagName) path.unshift(parent.tagName) ;
			parent = parent.parentElement || parent.parentNode ;
		}
		return path ;
	}
	this.formatPath	= function(e) {
		var label 	= "/" ;
		var path 	= this.getPath(e) ;
		for (n = 0 ; n < path.length ; n++) label += path[n] + "/" ;
		return label ;
	}
	this.pwd		= function() 	{ this.view.print(this.formatPath(this.current)) ; }
	this.ls			= function(opt) {
		this.manager.reset() ;
		var children = this.current.childNodes ;
		// Managing the simple "ls" command (default option)
		if (opt == null) {
			var lines = this.columns(children, 4, 64) ;
			for (n = 0 ; n < lines.length ; n++) this.view.print(lines[n]) ;
		// Managing all formatted "ls" commands (formatted output): see the format() function
		} else if (opt != "-d") {
			if (opt == "-dom") this.view.scroll("Element&nbsp;&nbsp;&nbsp;nodeName localName nodeType") ;
			if (opt != "-a") for (n = 0 ; n < children.length ; n++) {
				if (this.checkTag(children[n])) this.view.print( this.format(children[n], opt) ) ;	// Forwarding option to the format() method
			} else {
				for (n = 0 ; n < children.length ; n++) {
					switch(children[n].nodeType) {
					case 1	: 		this.view.print( this.format(children[n], "-l") ) ;
									break ;
					case 3	:		var value = children[n].nodeValue ;
									if (value == "\n") 	this.view.print( "TextNode&nbsp;&nbsp;&nbsp;CR") ;
									else 				this.view.print( "TextNode&nbsp;&nbsp;" + value) ;
									break ;
					default :		this.view.print( children[n].tagName + " (" + children[n].nodeType + ")" ) ;
					}
				}
			}
		// Managing not formatted "ls" command (option "-d")
		} else {
			if (opt == "-d") {
				var col		= 25 ;
				var dots	= "..." ;
				var prefix	= " (Type: " ;
				var suffix 	= ")" ;
				this.view.scroll( this.trailSpace("Current tag", col) 			+ this.current.tagName															) ;
				this.view.scroll( this.trailSpace("Node name", col)			 	+ this.current + prefix + this.current.nodeType + suffix						) ;
				this.view.scroll( this.trailSpace("Having attributes", col)		+ this.current.attributes + prefix + this.current.attributes.nodeType + suffix	) ;
				this.view.scroll( this.trailSpace("Containing childNodes", col)	+ children + prefix + children.nodeType + suffix								) ;
				this.view.scroll( this.trailSpace("DOM path", col)				+ this.formatPath(this.current)					 								) ;
				if ('innerText' in this.current) {
					this.view.scroll( this.trailSpace("innerText", col)			+ this.current.innerText.substring(0,80) + dots 								) ;
				}
				else {
					this.view.scroll( this.trailSpace("textContent", col)		+ this.current.textContent.substring(0,80) + dots 								) ;
				}				
			}
		}
		// Keep for testing or debug
		// this.manager.debug() ;
	}
	this.cd			= function(arg)	{
		var buffer = this.current ;
		if (( arg == null) || (arg == "-")) {
			if (arg == null)	this.current = this.home ;
			else				this.current = this.previous ;
			this.previous = buffer ;
			return ;
		}
		this.manager.reset() ;
		var	arg			= arg.toLowerCase() ;
		var children	= this.current.childNodes ;
		if (arg == "..") {
			var parent 	= this.current.parentElement || this.current.parentNode ;
			if (parent.tagName) this.current = parent ;
		} else for (n = 0 ; n < children.length ; n++) if (this.checkTag(children[n])) {
			var tag = children[n].tagName.toLowerCase() ;
			if (arg == tag) this.current = children[n] ;
			// Managing "virtual" tags, numbered by the TagManager.format() function
			else {
				var label = this.manager.format(children[n]).toLowerCase() ;
				if (arg == label) this.current = children[n] ;
			}
		}
		this.previous = buffer ;
	}
	// Verifies that the node has a name (text nodes doesn't)
	this.checkTag	= function(e) 		{ return e.tagName ; }
	this.format		= function(e, opt) {
		label 	= this.manager.format(e) ;
		label	= this.trailSpace(label, 10) ;
		if ((opt == "-l") || (opt == "-attr")) {
			for (k = 0 ; k < e.attributes.length ; k++) label += this.blank + e.attributes[k].name + ":" + e.attributes[k].value ;
			if (opt == "-l") label = label.substring(0, 80) + " ..." ;
		}
		// TODO: add other metadata to this list?
		if (opt == "-dom") label += this.trailSpace(e.nodeName, 10) + this.trailSpace(e.localName, 9) + this.trailSpace(e.nodeType, 10) ;
		return label ;
	}
	this.trailSpace	= function(s, n) {
		var counter	= s.length ;
		while (counter++ < n) { s += this.blank ; }
		return s ;
	}
	this.columns	= function(tokens, col, width) {
		if (tokens == null) return "empty" ;
		var	total	= 0 ;
		var	counter	= 0 ;
		var	temp 	= "" ;
		var	label 	= "" ;
		// Evaluating how many well defined elements are contained in the tokens array
		for (var i = 0 ; i < tokens.length ; i++) if (this.checkTag(tokens[i])) total++ ;
		var max 	= Math.ceil(total/col) ;
		var	space	= Math.ceil(width/col) ;
		var array	= new Array(max) ;
		var cols	= 0 ;
		total		= 0 ;
		for (var i = 0 ; i < tokens.length ; i++) if (this.checkTag(tokens[i])) {
			total++ ;
			label	= this.format(tokens[i], "-s") ;
			temp 	+= label ;
			cols 	+= label.length ;
			var x 	= total % col ;
			while (cols < space*x) {
				cols++ ;
				temp += this.blank ;
			}
			if (x == 0) {
				array[counter++] = temp ;
				cols = 0 ;
				temp = "" ;
			}
		}
		// Setting last element (if last line was not completed)
		if (x != 0) array[max-1] = temp ;
		return array ;
	}
}

/*************************************************************************************/
/***************	 MAIN JAVASCRIPT FUNCTIONS (not OOB)		**********************/
/*************************************************************************************/

// Initializing objects: this code-block does not involve HTML modification, hence
// can be run as soon as this library is available
function main(target) {
	viewObj 		= new ShellView("#", 20) ;
	handlerObj 		= new HandlerDOM(viewObj) ;
	controllerObj 	= new ShellController(viewObj) ;
	controllerObj.setHandler(handlerObj) ;
	handlerObj.setHome(target) ;
	viewObj.init() ;
	return viewObj ;
}
// Initializing HTML: can be run only when HTML markup is available
function activate(viewObj) {
	viewObj.getModel().clear() ;
	viewObj.getInput().focus() ;
	showTarget(targetURL) ;
}
// Setting target description (shell's caption)
function showTarget(target) {
	var span = document.getElementById("target_desc") ;
	if (target == null) span.innerHTML = document.URL ;		// Same as window.location.href ;
	else span.innerHTML = targetURL ;
}

/*************************************************************************************/
/***************	 LOAD/CHANGE NEW DOM STRUCTURE (not OOB)	**********************/
/*************************************************************************************/

// Change current DOM target, loading a new content from the input text
function loadFromText(viewObj) {
	openTextarea("Copy-paste HTML content in the textarea below") ;
}
// Set the new DOM content, using result of the "loadFromText()" functions
function setAreaContent(content) {
	setCurrentNode(content, false) ;
}
// Change current DOM target, loading a new content from the input URL
function loadFromUrl(viewObj) {
	var url = prompt("Insert new URL") ;
	// TODO: validate URL using a regular expression (shellDom 1.1 ???)
	if (url.length < 8) {
		var message  = "Wrong XML reference.\nPlease insert a valid URL pointing to the target DOM.\n" ;
		message	 	+= "Click OK to return to previous page (no change in current DOM navigation)" ;
		alert(message) ;
		viewObj.getInput().focus() ;
	// If we get here we assume we have a valid remote URL
	} else {
		xmlRequest = new XMLHttpRequest() ;
		xmlRequest.onreadystatechange = stateChanged ;
		xmlRequest.open("GET", url, true) ;
		xmlRequest.send(null) ;
		// Storing a reference in the global variable
		loadingURL = url ;
		//console.log("Started remote call to " + url) ;
	}
}	
// Callback for the XMLHttpRequest remote call
function stateChanged() {	
	// TODO: on remote servers, different from the server hosting this script, we get a response error
	// with status HTTP = 0. Why? Investigate the XMLHttpRequest protocol! (shellDom 1.1 ???)
	if (xmlRequest.readyState == 4) {
		if (xmlRequest.status == 200)	{
			// Keep for testing or debug
			//showRemoteContent(xmlRequest) ;
			setCurrentNode(xmlRequest.responseText, true) ;
		} else { 
			alert("Error receiveing the XML/HTML content") ; 
		}
	}
	//console.log("Ended remote call: ready = " + xmlRequest.readyState + " - status = " + xmlRequest.status) ;
}
// Load the input XML/HTML (in string format) into the current node reference
function setCurrentNode(text, remote) {
	var node = null ;
	try {
		// TODO: improve using DOMParser() ? (shellDom 1.1 ???)
		// Using just "innerHTML" we can not handle the <head> tag --> we must select only the HTML body !
		// Moreover, current sytem (below) does not support XML ! 
		var start 	= text.indexOf("body") ;
		var end 	= text.lastIndexOf("body") ;
		var body	= text.substring(start-1, end + 5) ;
		node = document.createElement("body") ;
		node.innerHTML = body ;
		if (remote) targetURL = loadingURL ;
		else 		targetURL = "INPUT TEXT" ;
	} catch(err) { 
		console.log("ERROR:" + err) ;
		// Restoring previous target
		node = shellDomObj.getModel().getHandler().getHome() ;
	}
	// Setting new node and restartig the shell view
	var handler = shellDomObj.getModel().getHandler() ;
	handler.setHome(node) ;
	activate(shellDomObj) ;
}
// Debug function
function showRemoteContent(request) {
	message	 = "XMLRequest text: " 		+ request.statusText + "\n\n" ;
	message	+= "XMLRequest headers:\n" 	+ request.getAllResponseHeaders() + "\n" ;
	message	+= "XMLRequest response:\n"	+ request.responseText ;
	alert(message) ;
}

/*************************************************************************************/
/***************	 MAIN CODE, EXECUTED WHEN PAGE IS LOADED	**********************/
/*************************************************************************************/

// Setting default target
var default_target 		= document.getElementsByTagName("body")[0] ;
// Keep for testing or debug
//var default_target 	= document.getElementById("dummy") ;
// var default_target 	= document.getElementsByTagName("tbody")[1] ;

// Running main() function and setting GLOBAL JavaScript variables
var areaContent	= null ;
var loadingURL	= null ;		// Last URL tried as "change target" operation (could be failed)
var targetURL	= null ;		// Current URL containing the current DOM tree (if null --> we'are still in the default DOM)
var xmlRequest	= null ;		// Global object used by the "change target" functions
var shellDomObj = main(default_target) ;

// Activating the shell (can be done when all the necessary markup has been loaded)
document.body.onload = function() {	activate(shellDomObj) ; }